@@ -172,7 +172,7 @@ static QString StringFromSelectedText(const SelectionText &selectedText)
 	}
 }
 
-static void AddRectangularToMime(QMimeData *mimeData, [[maybe_unused]] QString su)
+static void AddRectangularToMime(QMimeData *mimeData, [[maybe_unused]] const QString &su)
 {
 #if defined(Q_OS_WIN)
 	// Add an empty marker
@@ -551,6 +551,7 @@ QByteArray ScintillaQt::BytesForDocument(const QString &text) const
 	}
 }
 
+namespace {
 
 class CaseFolderDBCS : public CaseFolderTable {
 	QTextCodec *codec;
@@ -578,6 +579,8 @@ public:
 	}
 };
 
+}
+
 std::unique_ptr<CaseFolder> ScintillaQt::CaseFolderForEncoding()
 {
 	if (pdoc->dbcsCodePage == SC_CP_UTF8) {
