@@ -567,6 +567,8 @@ static const char *propertiesToForward[] = {
 	"lexer.nim.raw.strings.highlight.ident",
 	"lexer.pascal.smart.highlighting",
 	"lexer.props.allow.initial.spaces",
+	"lexer.python.decorator.attributes",
+	"lexer.python.identifier.attributes",
 	"lexer.python.keywords2.no.sub.identifiers",
 	"lexer.python.literals.binary",
 	"lexer.python.strings.b",
@@ -1185,6 +1187,8 @@ void SciTEBase::ReadProperties() {
 	wOutput.SetLayoutCache(static_cast<SA::LineCache>(
 				       props.GetInt("output.cache.layout", static_cast<int>(SA::LineCache::Caret))));
 
+	wEditor.SetLayoutThreads(props.GetInt("threads.layout", 1));
+
 	bracesCheck = props.GetInt("braces.check");
 	bracesSloppy = props.GetInt("braces.sloppy");
 
